import os

import pytest

from tktl.login import login, logout

pytest_plugins = [
    "tests.cli",
    "tests.sdk",
    "tests.clients",
    "tests.managers",
    "tests.commands",
]


@pytest.fixture
def user_key():
    return os.environ["TEST_USER"], os.environ["TEST_USER_API_KEY"]


@pytest.fixture
def test_user_repos():
    return (
        "tktl-admin/sample-project",
        "tktl-admin/other-project",
        "tktl-admin/integ-testing",
    )


@pytest.fixture(scope="module")
def logged_in_context():
    yield login(os.environ["TEST_USER_API_KEY"])
    logout()



import json
from functools import wraps

import requests

from tktl.core.config import settings
from tktl.core.exceptions import InvalidInputError
from tktl.core.loggers import LOG
from tktl.core.managers.auth import AuthConfigManager
from tktl.core.strings import CLIStrings, HeaderStrings

UNAUTHORIZED_EXTENDED_INFO = (
    "\n\nNote: Please keep in mind that currently you can login only with the email and "
    "password from your Paperspace account. If you're using AD, SAML or GitHub "
    "credentials, please log into the Paperspace Console and create an API key for use with "
    "the CLI client. For more information, please refer to the CLI client documentation."
)


def is_error_or_missing_keys_print(res, *required_keys):
    if "error" in res:
        if "message" in res:
            print(res["message"])
            return True
        if "message" in res["error"]:
            error_message = res["error"]["message"] + UNAUTHORIZED_EXTENDED_INFO
            print(error_message)
            return True
        print(json.dumps(res, indent=2, sort_keys=True))
        return True
    elif not all(key in res for key in required_keys):
        if "message" in res:
            print(res["message"])
            return True
        print(json.dumps(res, indent=2, sort_keys=True))
        return True
    return False


def login(api_key=None):
    """

    Parameters
    ----------
    api_key

    Returns
    -------

    """

    if not api_key:
        raise InvalidInputError(CLIStrings.INVALID_INPUT + "api_key must be provided")

    AuthConfigManager.set_api_key(api_key)
    return True


def set_api_key(api_key):
    AuthConfigManager.set_api_key(api_key=api_key)
    path = (
        f"{AuthConfigManager.get_tktl_config_path()}/"
        f"{AuthConfigManager.SETTINGS.CONFIG_FILE_NAME}"
    )
    LOG.log(f"Successfully added your API Key to {path} You're ready to go!")
    return True


def logout():
    AuthConfigManager.clear_tktl_config()
    return True


def validate_key(suppress=False, api_key=None):
    if not api_key:
        api_key = AuthConfigManager.get_api_key()
    if not api_key:
        LOG.error(CLIStrings.AUTH_ERROR_MISSING_KEY)
        return False

    r = requests.get(
        f"{settings.TAKTILE_API_URL}{settings.API_V1_STR}/users/me",
        headers={"Accept": HeaderStrings.APPLICATION_JSON, "X-Api-Key": api_key},
    )
    try:
        r.raise_for_status()
    except requests.exceptions.RequestException as e:
        LOG.error(f'Request failed: {r.json()["detail"]}')
        return False

    if not suppress:
        LOG.log("Login successful!", color="green")
    return True


def validate_decorator(func):
    @wraps(func)
    def wrapped_validation(*args, **kwargs):
        if not validate_key(suppress=True):
            return
        return func(*args, **kwargs)

    return wrapped_validation



import os

from tktl.core.clients.arrow import ArrowFlightClient
from tktl.core.managers.auth import AuthConfigManager


def test_instantiate_client():

    key = os.environ["TEST_USER_API_KEY"]
    AuthConfigManager.set_api_key(key)

    client = ArrowFlightClient(
        api_key=key, repository_name=f"{os.environ['TEST_USER']}/test-new"
    )
    assert client.location is None



from typing import Any, Callable, Dict, List, Optional, Union

from pydantic import UUID4

from .. import loggers as sdk_logger
from ..exceptions import TaktileSdkError
from ..schemas.repository import Endpoint, Repository, RepositoryDeployment
from ..t import Resources

_logger = sdk_logger.SdkLogger()

default_comp = lambda x, y: x == y


def filter_prop(
    resources: List[Resources], prop_name: str, value: Any, key: Callable = default_comp
):
    if value is not None:
        return [d for d in resources if key(value, getattr(d, prop_name))]
    return resources


def filter_deployments(
    deployments: List[RepositoryDeployment],
    deployment_id: UUID4,
    commit_sha: str = None,
    branch_name: str = None,
    status_name: str = None,
) -> List[RepositoryDeployment]:
    if deployment_id:
        deployments = [d for d in deployments if d.id == deployment_id]

        if deployments:
            return deployments
        else:
            _logger.warning("No deployments with matching id found")

    for n, v in zip(
        ["commit_hash", "branch_name", "status"], [commit_sha, branch_name, status_name]
    ):
        if n == "commit_hash" and commit_sha:
            if len(commit_sha) < 40:
                comp = lambda x, y: y.startswith(x)
            else:
                comp = default_comp
        else:
            comp = default_comp
        deployments = filter_prop(deployments, prop_name=n, value=v, key=comp)

    return deployments


def filter_repositories(
    repositories: List[Repository],
    repository_id: UUID4,
    repository_name: str = None,
    repository_owner: str = None,
) -> List[Repository]:
    if repository_id:
        repositories = [r for r in repositories if r.id == repository_id]
        if repositories:
            return repositories
        else:
            _logger.warning("No repositories with matching id found")

    for n, v in zip(
        ["repository_name", "repository_owner"], [repository_name, repository_owner]
    ):
        repositories = filter_prop(repositories, prop_name=n, value=v)
    return repositories


def filter_endpoints(
    endpoints: List[Endpoint], endpoint_id: UUID4, endpoint_name: str = None
) -> List[Endpoint]:
    if endpoint_id:
        endpoints = [r for r in endpoints if r.id == endpoint_id]
        if endpoints:
            return endpoints
        else:
            _logger.warning("No endpoints with matching id found")

    for n, v in zip(["endpoint_name"], [endpoint_name]):
        endpoints = filter_prop(endpoints, prop_name=n, value=v)
    return endpoints


def get_deployment_from_endpoint_and_branch(
    endpoint_mapping: List[Dict[str, Union[Endpoint, RepositoryDeployment]]],
    endpoint_name: str,
    branch_name: str = None,
) -> Optional[RepositoryDeployment]:
    filtered = [e for e in endpoint_mapping if e if e["endpoint"].name == endpoint_name]
    if len(filtered) == 0:
        _logger.warning(
            f"No endpoints with name: {endpoint_name} found across all deployed branches"
        )
        return
    if len(filtered) == 1:
        deployment = filtered[0]["deployment"]
        if branch_name is not None and deployment.branch_name != branch_name:
            _logger.warning(
                f"No endpoint {endpoint_name} for branch {branch_name}, but found one for branch: "
                f"{deployment.branch_name}. Returning that instead"
            )
        return deployment
    else:
        available_branches = set(
            [e["deployment"].branch_name for e in endpoint_mapping]
        )
        if branch_name is None:
            _logger.warning(
                f"Ambiguous: more than one branch for repository with endpoint: {endpoint_name}. Please "
                f"specify a branch. Available branches: with such endpoint: {', '.join(available_branches)}"
            )
            return
        branch_filtered = [
            e for e in filtered if e["deployment"].branch_name == branch_name
        ]
        if len(branch_filtered) == 0:
            _logger.warning(
                f"No endpoints with name: {endpoint_name} and branch {branch_name} found. Available "
                f"branches with such endpoint: {', '.join(available_branches)}"
            )
            return
        elif len(branch_filtered) == 1:
            return branch_filtered[0]["deployment"]
        else:
            print(
                [
                    (e["deployment"].branch_name, e["deployment"].status)
                    for e in endpoint_mapping
                ]
            )
            # Impossible to have more than one endpoint in the same branch for the same repo with the same name
            raise TaktileSdkError("This should never, ever happen")



from pyarrow.flight import (
    ClientAuthHandler,
    FlightClient,
    FlightDescriptor,
    FlightInfo,
    Ticket,
)

from tktl.core.clients import Client
from tktl.core.config import settings
from tktl.core.exceptions import AuthenticationError
from tktl.core.schemas.repository import _format_grpc_url
from tktl.core.serializers import deserialize_arrow, serialize_arrow
from tktl.core.t import ServiceT


class ApiKeyClientAuthHandler(ClientAuthHandler):
    """An example implementation of authentication via ApiKey."""

    def __init__(self, api_key: str):
        super(ApiKeyClientAuthHandler, self).__init__()
        self.api_key = api_key

    def authenticate(self, outgoing, incoming):
        outgoing.write(self.api_key)
        self.api_key = incoming.read()

    def get_token(self):
        return self.api_key


class ArrowFlightClient(Client):

    TRANSPORT = ServiceT.GRPC

    def __init__(
        self,
        api_key: str,
        repository_name: str,
        branch_name: str = None,
        local: bool = False,
    ):
        super().__init__(
            api_key=api_key,
            repository_name=repository_name,
            branch_name=branch_name,
            local=local,
        )

    @staticmethod
    def format_url(url: str) -> str:
        return _format_grpc_url(url)

    @property
    def local_endpoint(self):
        return settings.LOCAL_ARROW_ENDPOINT

    def list_deployments(self):
        pass

    def list_commands(self):
        return self.client.list_actions()

    def authenticate(self, endpoint_name: str):
        location = self.get_endpoint_and_location(endpoint_name)
        if not location:
            return
        self.location = location
        client = FlightClient(location=self.location)
        client.authenticate(
            ApiKeyClientAuthHandler(api_key=self.taktile_client.api_key)
        )
        self.client = client

    def predict(self, inputs):
        table = serialize_arrow(inputs)
        descriptor = self.get_flight_info(command_name=str.encode(self.endpoint))
        writer, reader = self.client.do_exchange(descriptor.descriptor)
        with writer:
            writer.begin(table.schema)
            writer.write_table(table)
            writer.done_writing()
            table = reader.read_all()
            return deserialize_arrow(table)

    def get_sample_data(self):
        if not self.endpoint:
            raise AuthenticationError(
                "Please authenticate against a specific endpoint first"
            )
        x_info = self.client.do_get(Ticket(ticket=str.encode(f"{self.endpoint}__X")))
        y_info = self.client.do_get(Ticket(ticket=str.encode(f"{self.endpoint}__y")))
        return (
            deserialize_arrow(x_info.read_all()),
            deserialize_arrow(y_info.read_all()),
        )

    def get_schema(self):
        if not self.endpoint:
            raise AuthenticationError(
                "Please authenticate against a specific endpoint first"
            )
        info = self.get_flight_info(str.encode(self.endpoint))
        return info.schema

    def get_flight_info(self, command_name: bytes) -> FlightInfo:
        descriptor = FlightDescriptor.for_command(command_name)
        return self.client.get_flight_info(descriptor)



import abc

from tktl.core.clients import TaktileClient
from tktl.core.config import settings
from tktl.core.loggers import LOG


class CommandBase(object):
    def __init__(self, api=None):
        self.api = api


class BaseApiCommand:

    __metaclass__ = abc.ABCMeta

    def __init__(
        self, api_url=settings.DEPLOYMENT_API_URL,
    ):
        self.api_url = api_url
        self.client = self._get_client()

    @abc.abstractmethod
    def execute(self, *args, **kwargs):
        pass

    def _get_client(self):
        return TaktileClient(api_url=self.api_url)


class BaseDeploymentApiCommand(BaseApiCommand):
    def __init__(self):
        super().__init__(api_url=settings.DEPLOYMENT_API_URL)

    def execute(self, *args, **kwargs):
        raise NotImplementedError


class BaseTaktileApiCommand(BaseApiCommand):
    def __init__(self):
        super().__init__(api_url=settings.TAKTILE_API_URL)

    def execute(self, *args, **kwargs):
        raise NotImplementedError



import pytest
from pydantic import UUID4

from tktl.commands.deployments import GetDeployments, GetEndpoints
from tktl.core.exceptions.exceptions import APIClientException
from tktl.login import logout


def test_get_deployment_commands(logged_in_context):
    cmd = GetDeployments()
    result = cmd.execute(UUID4("b15e5dd1-fbab-42bd-a646-f50d22b6b425"), "", "", "")
    assert len(result) == 1

    result = cmd.execute("", "", "", "", return_all=True)
    assert len(result) >= 3

    cmd = GetEndpoints()
    result = cmd.execute(UUID4("b15e5dd1-fbab-42bd-a646-f50d22b6b425"), "")
    assert len(result) == 0

    cmd = GetDeployments()
    with pytest.raises(APIClientException) as e:
        logout()
        cmd.execute(UUID4("b15e5dd1-fbab-42bd-a646-f50d22b6b425"), "", "", "")



import functools
import warnings
from abc import ABC, abstractmethod
from typing import Any, Callable, List, Optional, Tuple, Union

import pandas

from tktl.core.exceptions import exceptions
from tktl.core.loggers import LOG
from tktl.core.t import EndpointKinds
from tktl.registration.schema import EndpointInputSchema, EndpointOutputSchema
from tktl.registration.validation import (
    validate_binary,
    validate_outputs,
    validate_shapes,
)


class Endpoint(ABC):
    KIND: str
    input_schema: EndpointInputSchema
    output_schema: EndpointOutputSchema

    def __init__(self, func: Callable, profile_columns: Optional[List] = None):
        self._func = func
        self.profile_columns = profile_columns

    @property
    def func(self):
        return self._func

    @property
    def series_func(self) -> Union[Any, pandas.Series]:
        name = self._func.__name__

        def to_series(x):
            return pandas.Series(self.func(x))

        to_series.__name__ = name
        return to_series

    @property
    @abstractmethod
    def profiling_supported(self):
        return True

    @abstractmethod
    def get_input_and_output_for_profiling(
        self, full_x=False, reset_index=False
    ) -> Tuple[pandas.DataFrame, pandas.Series]:
        raise NotImplementedError

    @property
    def input_names(self):
        return []

    @property
    def output_names(self):
        return []


class TabularEndpoint(Endpoint):
    explain_input_schema: EndpointInputSchema
    KIND: str = EndpointKinds.TABULAR

    def __init__(self, func, X, y, profile_columns):
        super().__init__(func, profile_columns)
        self.input_schema = EndpointInputSchema(
            value=X, endpoint_kind=self.KIND, endpoint_name=self.func.__name__
        )

        self.output_schema = EndpointOutputSchema(
            value=y, endpoint_kind=self.KIND, endpoint_name=self.func.__name__
        )
        if self.profiling_supported:
            input_value = self.input_schema.get_pandas_representation()
            if not profile_columns:
                profile_columns = input_value.columns

            self.explain_input_schema = EndpointInputSchema(
                value=input_value[profile_columns],
                endpoint_kind=self.KIND,
                endpoint_name=f"explain__{self.func.__name__}",
            )

    def _profiling_supported(self) -> Union[bool, pandas.Series]:
        if (
            not self.input_schema.pandas_convertible
            and not self.output_schema.pandas_convertible
        ):

            return False
        try:
            predictions = self.func(self.input_schema.value)
        except Exception as e:
            raise exceptions.ValidationException(
                f"Function provided is unable to produce predictions from given sample values: {e}"
            )
        return predictions

    @property
    def profiling_supported(self):
        predictions = self._profiling_supported()
        if predictions is False:
            return False
        else:
            try:
                x_frame = self.input_schema.get_pandas_representation()
                y_series = self.output_schema.get_pandas_representation()
            except ValueError:
                return False
            return validate_outputs(predictions) and validate_shapes(
                x_frame, y_series, predictions
            )

    @property
    def input_names(self):
        return self.input_schema.names

    @property
    def explain_input_names(self):
        return self.explain_input_schema.names

    @property
    def output_names(self):
        return self.output_schema.names

    def get_input_and_output_for_profiling(
        self, full_x=False, reset_index=False
    ) -> Tuple[pandas.DataFrame, pandas.Series]:
        return self._drop_missing_values(reset_index=reset_index, full_x=full_x)

    def _drop_missing_values(
        self, full_x=False, reset_index=False
    ) -> Tuple[pandas.DataFrame, pandas.Series]:
        if full_x:
            x = self.input_schema.get_pandas_representation(reset_index=reset_index)
        else:
            x = self.explain_input_schema.get_pandas_representation(
                reset_index=reset_index
            )

        y = self.output_schema.get_pandas_representation(reset_index=reset_index)
        not_missing = [i for i, v in enumerate(y) if not pandas.isna(v)]
        n_missing = len(y) - len(not_missing)
        if n_missing > 0:
            warnings.warn(f"y contains {n_missing} missing values that will be dropped")
            return x.iloc[not_missing], y.iloc[not_missing]
        else:
            return x, y


class BinaryEndpoint(TabularEndpoint):
    KIND = EndpointKinds.BINARY

    def __init__(self, func, X, y, profile_columns):
        super().__init__(func, X, y, profile_columns)

    @property
    def profiling_supported(self):
        predictions = self._profiling_supported()
        if predictions is False:
            return False
        return super().profiling_supported and validate_binary(predictions)


class RegressionEndpoint(TabularEndpoint):
    KIND = EndpointKinds.REGRESSION

    def __init__(self, func, X, y, profile_columns):
        super().__init__(func, X, y, profile_columns)

    @property
    def profiling_supported(self):
        predictions = self._profiling_supported()
        if predictions is False:
            return False
        return super().profiling_supported and validate_outputs(predictions)


class CustomEndpoint(Endpoint):
    def get_input_and_output_for_profiling(
        self, **kwargs
    ) -> Tuple[pandas.DataFrame, pandas.Series]:
        raise ValueError("Custom endpoint does not have profiling enabled")

    KIND = EndpointKinds.CUSTOM

    def __init__(self, func, payload_model=None, response_model=None, X=None, y=None):
        super().__init__(func)
        try:
            assert isinstance(func, Callable)
        except AssertionError:
            raise exceptions.ValidationException("Endpoint function is not a callable")

        models_defined = payload_model and response_model
        x_and_y_defined = (X is not None) and (y is not None)
        if not models_defined and not x_and_y_defined:
            raise ValueError(
                "For custom endpoints, either define sample data or payload models"
            )

        if models_defined and x_and_y_defined:
            raise ValueError(
                "For custom endpoints, either define sample data or payload models"
            )

        self.input_schema = EndpointInputSchema(
            value=X,
            endpoint_kind=self.KIND,
            endpoint_name=self.func.__name__,
            user_defined_model=payload_model,
        )
        self.output_schema = EndpointOutputSchema(
            value=y,
            endpoint_kind=self.KIND,
            endpoint_name=self.func.__name__,
            user_defined_model=response_model,
        )

    @property
    def profiling_supported(self):
        return False


class Tktl:
    def __init__(self):
        self.endpoints = []

    # This is the user-facing decorator for function registration
    def endpoint(
        self,
        func: Callable = None,
        kind: str = EndpointKinds.REGRESSION,
        X: Any = None,
        y: Any = None,
        payload_model=None,
        response_model=None,
        profile_columns: Optional[List] = None,
    ):
        """Register function as a Taktile endpoint

        Parameters
        ----------
        func : Callable, optional
            Function that describes the desired operation, by default None
        kind : str, optional
            Specification of endpoint type ("regression", "binary", "custom"),
            by default "regression"
        X : pd.DataFrame, optional
            Reference input dataset for testing func. Used when argument "kind"
            is set to "regression" or "binary", by default None.
        y : pd.Series, optional
            Reference output for evaluating func. Used when argument "kind"
            is set to "regression" or "binary", by default None.
        payload_model:
            Type hint used for documenting and validating payload. Used in
            custom endpoints only.
        response_model:
            Type hint used for documenting and validating response. Used in
            custom endpoints only.
        profile_columns: List, optional
            List of column names for which to run profiling. Used in binary
            and regression endpoints only.

        Returns
        -------
        Callable
            Wrapped function
        """
        if func is None:
            return functools.partial(
                self.endpoint,
                kind=kind,
                X=X,
                y=y,
                payload_model=payload_model,
                response_model=response_model,
                profile_columns=profile_columns,
            )

        if kind == "tabular":
            endpoint = TabularEndpoint(
                func=func, X=X, y=y, profile_columns=profile_columns
            )
        elif kind == "regression":
            endpoint = RegressionEndpoint(
                func=func, X=X, y=y, profile_columns=profile_columns
            )
        elif kind == "binary":
            endpoint = BinaryEndpoint(
                func=func, X=X, y=y, profile_columns=profile_columns
            )
        elif kind == "custom":
            endpoint = CustomEndpoint(
                func=func,
                payload_model=payload_model,
                response_model=response_model,
                X=X,
                y=y,
            )
        else:
            raise exceptions.ValidationException(f"Unknown endpoint kind: '{kind}'")

        self.endpoints.append(endpoint)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            pred = func(*args, **kwargs)
            return pred

        return wrapper



from collections.abc import Sequence
from functools import singledispatch
from typing import List, Type, Union

import numpy
import pandas
from pydantic import BaseModel

from tktl.core.loggers import LOG
from tktl.core.serializers import rest
from tktl.core.serializers.base import CustomDeserializingModelT


def deserialize_rest(
    model_input: Union[List[CustomDeserializingModelT], CustomDeserializingModelT]
):
    if isinstance(model_input, list):
        return [v.deserialize() for v in model_input]
    else:
        return model_input.deserialize()


@singledispatch
def serialize_rest(model_input, output_model: Type[BaseModel]):
    return output_model(**{"value": model_input})


@serialize_rest.register
def _(model_input: BaseModel, output_model: Type[BaseModel]):
    return rest.PassThroughSerializer.serialize(
        value=model_input, output_model=output_model
    )


@serialize_rest.register
def _(model_input: Sequence, output_model: Type[BaseModel]):
    return rest.SequenceSerializer.serialize(
        value=model_input, output_model=output_model
    )


@serialize_rest.register(dict)
def _(model_input, output_model: Type[BaseModel]):
    return rest.SequenceSerializer.serialize(
        value=model_input, output_model=output_model
    )


@serialize_rest.register
def _(model_input: numpy.ndarray, output_model: Type[BaseModel]):
    return rest.ArraySerializer.serialize(value=model_input, output_model=output_model)


@serialize_rest.register
def _(model_input: pandas.DataFrame, output_model: Type[BaseModel]):
    return rest.DataFrameSerializer.serialize(
        value=model_input, output_model=output_model
    )


@serialize_rest.register
def _(model_input: pandas.Series, output_model: Type[BaseModel]):
    return rest.SeriesSerializer.serialize(value=model_input, output_model=output_model)



import json
import os

from openapi_schema_pydantic import OpenAPI

from tktl.core.clients import Client
from tktl.core.clients.http_client import API
from tktl.core.config import settings
from tktl.core.schemas.repository import _format_http_url
from tktl.core.t import ServiceT


class RestClient(Client):
    def get_schema(self, *args, **kwargs):
        pass

    TRANSPORT = ServiceT.REST

    def __init__(self, api_key: str, repository_name: str, local: bool = False):
        super().__init__(api_key=api_key, repository_name=repository_name)

        self.local = local

    def predict(self, inputs):
        response = self.client.post(
            url=f"model/{self.endpoint}", data=json.dumps(inputs)
        )
        return response.json()

    @staticmethod
    def format_url(url: str):
        return _format_http_url(url)

    def authenticate(self, endpoint_name: str):
        location = self.get_endpoint_and_location(endpoint_name)
        if not location:
            return
        self.location = location
        client = API(api_url=self.location)
        self.client = client

    @property
    def local_endpoint(self):
        return settings.LOCAL_REST_ENDPOINT

    def list_deployments(self):
        pass

    def get_sample_data(self):
        schema = self.client.get(url="openapi.json")
        openapi = OpenAPI.parse_obj(schema.json())
        request_reference, response_reference = get_endpoint_model_reference(
            openapi, endpoint=self.endpoint
        )
        sample_input = openapi.components.schemas[request_reference].example
        if not sample_input:
            self.logger.warning("No sample input found for endpoint")
        sample_output = openapi.components.schemas[response_reference].example
        if not sample_output:
            self.logger.warning("No sample output found for endpoint")
        return sample_input, sample_output


def get_endpoint_model_reference(openapi: OpenAPI, endpoint: str):
    for k in openapi.paths.keys():
        if f"/model/{endpoint}" == k:
            request_ref = (
                openapi.paths[k]
                .post.requestBody.content["application/json"]
                .media_type_schema.ref
            )
            response_ref = (
                openapi.paths[k]
                .post.responses["200"]
                .content["application/json"]
                .media_type_schema.ref
            )
            return os.path.basename(request_ref), os.path.basename(response_ref)
    return None
